import io
import unittest

from cardutil.mciipm import VbsWriter, VbsReader, change_encoding, change_param_encoding, IpmReader, IpmWriter

message_ebcdic_raw = (
        '1144'.encode('cp500') +
        b'\xF0\x10\x05\x42\x84\x61\x80\x02\x02\x00\x00\x04\x00\x00\x00\x00' +
        ('164444555544445555111111000000009999201508151715123456789012333123423579957991200000'
         '012306120612345612345657994211111111145BIG BOBS\\70 FERNDALE ST\\ANNERLEY\\4103  QLD'
         'AUS0080001001Y99901600000000000000011234567806999999').encode('cp500'))

message_ascii_raw = (
        b'1144' +
        b'\xF0\x10\x05\x42\x84\x61\x80\x02\x02\x00\x00\x04\x00\x00\x00\x00' +
        b'164444555544445555111111000000009999201508151715123456789012333123423579957991200000'
        b'012306120612345612345657994211111111145BIG BOBS\\70 FERNDALE ST\\ANNERLEY\\4103  QLD'
        b'AUS0080001001Y99901600000000000000011234567806999999')


class MciIpmTestCase(unittest.TestCase):

    def test_real_message_example_ascii(self):
        # create the input ipm file bytes -- test_file
        message_list = [message_ascii_raw for _ in range(5)]
        with io.BytesIO() as in_data:

            # write vbs test file
            writer = VbsWriter(in_data)
            for message in message_list:
                writer.write(message)
            else:
                writer.close()
            print_stream(in_data, "VBS in data")

            # read vbs test file
            reader = IpmReader(in_data)
            results = list(reader)
        self.assertEqual(len(results), len(message_list))

    def test_real_message_example_ebcdic(self):
        message_list = [message_ebcdic_raw for _ in range(5)]
        with io.BytesIO() as in_data:
            # write 1014 blocked test file
            writer = VbsWriter(in_data, blocked=True)
            for message in message_list:
                writer.write(message)
            else:
                writer.close()
            print_stream(in_data, "1014 blocked in data")

            # read blocked test file
            reader = IpmReader(in_data, encoding='cp500', blocked=True)
            results = list(reader)
            print(results)

        self.assertEqual(len(results), len(message_list))

    def test_ipmwriter_vbs_file(self):
        record = {'MTI': '1111', 'DE2': '8888999988889999'}
        records = [record for _ in range(5)]

        with io.BytesIO() as out_data:
            writer = IpmWriter(out_data)
            for record in records:
                writer.write(record)
            writer.close()

            print_stream(out_data, 'VBS output file')

            reader = IpmReader(out_data)
            results = list(reader)
            print(results)

        self.assertEqual(results, records)

    def test_ipmwriter_blocked_file(self):
        record = {'MTI': '1111', 'DE2': '8888999988889999'}
        records = [record for _ in range(5)]

        with io.BytesIO() as out_data:
            writer = IpmWriter(out_data, blocked=True)
            for record in records:
                writer.write(record)
            writer.close()

            print_stream(out_data, 'VBS output file')

            reader = IpmReader(out_data, blocked=True)
            results = list(reader)
            print(results)

        self.assertEqual(results, records)

    def test_vbsreader_vbs_file(self):
        # create the input file bytes -- test_file
        records = [b'12345678901234567890' for _ in range(5)]
        with io.BytesIO() as in_data:

            # write vbs test file
            writer = VbsWriter(in_data)
            for record in records:
                writer.write(record)
            else:
                writer.close()
            print_stream(in_data, "VBS in data")

            reader = VbsReader(in_data)
            results = list(reader)
            print(results)

        self.assertEqual(results, records)

    def test_vbsreader_blocked_file(self):
        # create the input file bytes -- test_file
        records = [b'12345678901234567890' for _ in range(5)]
        with io.BytesIO() as in_data:

            # write vbs test file
            writer = VbsWriter(in_data, blocked=True)
            for record in records:
                writer.write(record)
            else:
                writer.close()
            print_stream(in_data, "Blocked vbs data")

            reader = VbsReader(in_data, blocked=True)
            results = list(reader)
            print(results)
            self.assertEqual(results, records)

    def test_change_encoding(self):
        # add 5 records to a list
        message_list = [message_ascii_raw for _ in range(1)]

        # create test file
        vbs_in = io.BytesIO()
        writer = VbsWriter(vbs_in, blocked=True)
        for message in message_list:
            writer.write(message)
        writer.close()

        # print it
        print_stream(vbs_in, "Blocked in data")

        # process the encoding
        param_out = io.BytesIO()
        change_param_encoding(vbs_in, param_out, in_encoding='latin1', out_encoding='latin1')
        print_stream(param_out, "Change param encoding")

        vbs_in.seek(0)
        ipm_out = io.BytesIO()
        change_encoding(vbs_in, ipm_out, in_encoding='latin1', out_encoding='latin1')
        print_stream(ipm_out, "Change encoding")

        vbs_in.seek(0)
        vbs_in_value = vbs_in.read()

        param_out.seek(0)
        param_out_value = param_out.read()
        self.assertEqual(vbs_in_value, param_out_value)

        ipm_out.seek(0)
        ipm_out_value = ipm_out.read()
        self.assertEqual(vbs_in_value, ipm_out_value)


def print_stream(stream, description):
    stream.seek(0)
    data = stream.read()
    print(description)
    print(data)
    stream.seek(0)


if __name__ == '__main__':
    unittest.main()



import binascii
from array import array


class BitArray:
    """
    This is a minimal native python replacement for the bitarray module that was used to interpret
    the file bitmap. The bitarray module is written in c and does not provide binary wheels so
    forces users to have compilers installed just to install.
    This small class provides the required functions from that library.
    """
    endian = 'big'
    bytes = b''

    def __init__(self, endian='big'):
        self.endian = endian

    def frombytes(self, array_bytes):
        self.bytes = array_bytes

    def tobytes(self):
        return self.bytes

    def tolist(self):
        swap_bytes = array('B', self.bytes)
        if self.endian == 'little':
            for i, n in enumerate(swap_bytes):
                swap_bytes[i] = int('{:08b}'.format(n)[::-1], 2)
        width = len(self.bytes)*8
        try:
            swapped_bytes = swap_bytes.tobytes()
        except AttributeError:  # 2.7 does not recognise .tobytes method. 2.6 does!
            swapped_bytes = swap_bytes.tostring()
        bit_list = '{bytes:0{width}b}'.format(bytes=int(binascii.hexlify(swapped_bytes), 16), width=width)
        return [bit == '1' for bit in bit_list]

    def fromlist(self, bytelist):
        # https://stackoverflow.com/questions/32675679/convert-binary-string-to-bytearray-in-python-3
        binary_value = ''.join(['1' if val else '0' for val in bytelist])
        self.bytes = self.bitstring_to_bytes(binary_value)
        # Python 3 only..
        # self.bytes = int(binary_value, 2).to_bytes(len(binary_value) // 8, byteorder='big')

    @staticmethod
    def bitstring_to_bytes(s):
        v = int(s, 2)
        b = bytearray()
        while v:
            b.append(v & 0xff)
            v >>= 8
        return bytes(b[::-1])



"""
The config module contains the default configuration for the cardutil library

ISO8583 message configuration
=============================

The iso8583 module functions require configuration that details how to process the bitmap in a message.

Field definitions in an ISO8583 message are generally the same but there can be variations you may need to cater for
with the different card schemes. The library supports provision of custom ISO8583 configuration.

If a configuration is not provided, the variable ``config['bit_config']`` in ``config.py`` provides a Mastercard
specific ISO8583 configuration.

The config is in the form of a python dictionary structured as follows::

    bit_config = {
        "1": {
            "field_name": "Bitmap secondary",
            "field_type": "FIXED",
            "field_length": 8},
        "2": {
            "field_name": "PAN",
            "field_type": "LLVAR",
            "field_length": 0
            "field_processor": "PAN",
            "field_python_type": "string"
            }
    }

The config dictionary contains a string key for each valid bit in the bitmap.
For example, the config for bit 1 with have the key "1".

The value for the bit is a dictionary containing the following values:

* ``field_name`` : description of the bit field
* ``field_type``: defines the ISO8583 field type
   * ``FIXED`` : field with fixed length
   * ``LLVAR`` : field with variable length - 2 character length value
   * ``LLLVAR`` : field with variable length - 2 character length value
* ``field_length``: length of the field. Use zero for variable length fields.
* ``field_processor`` (optional): a process to apply to a field.
   * ``PAN``: For use with PAN fields. Mask PAN using first 6, last 4 pattern
   * ``PAN-PREFIX``: For use with PAN fields. Only get first 9 PAN numbers -- prefix
   * ``ICC``: Mastercard ICC field. Adds TAGxxxx keys to output
   * ``PDS``: Mastercard PDS field. Processes PDSxxxx fields
   * ``DE43``: Mastercard Merchant details field. Adds the following subfields to output
      * DE43_NAME,
      * DE43_ADDRESS
      * DE43_SUBURB
      * DE43_POSTCODE
      * DE43_STATE
      * DE43_COUNTRY fields
* ``field_python_type`` (optional) : when processing between iso and python, determines the python object type.
    * ``string``: default if no type provided
    * ``int``
    * ``decimal``: use if your field has decimal place
    * ``datetime``: Uses the follow python date pattern ``"%y%m%d%H%M%S"``
"""

config = {
    "bit_config": {
        "1": {"field_name": "Bitmap secondary", "field_type": "FIXED", "field_length": 8},
        "2": {"field_name": "PAN", "field_type": "LLVAR", "field_length": 0},  # "field_processor": "PAN"},
        "3": {"field_name": "Processing code", "field_type": "FIXED", "field_length": 6},
        "4": {"field_name": "Amount transaction", "field_type": "FIXED", "field_length": 12,
              "field_python_type": "long"},
        "5": {"field_name": "Amount, Reconciliation", "field_type": "FIXED", "field_length": 12,
              "field_python_type": "long"},
        "6": {"field_name": "Amount, Cardholder billing", "field_type": "FIXED", "field_length": 12,
              "field_python_type": "long"},
        "9": {"field_name": "Conversion rate, Reconciliation", "field_type": "FIXED", "field_length": 8,
              "field_python_type": "long"},
        "10": {"field_name": "Conversion rate, Cardholder billing", "field_type": "FIXED", "field_length": 8,
               "field_python_type": "long"},
        "12": {"field_name": "Date/Time local transaction", "field_type": "FIXED", "field_length": 12,
               "field_python_type": "datetime"},
        "14": {"field_name": "Expiration date", "field_type": "FIXED", "field_length": 4},
        "22": {"field_name": "Point of service data code", "field_type": "FIXED", "field_length": 12},
        "23": {"field_name": "Card sequence number", "field_type": "FIXED", "field_length": 3},
        "24": {"field_name": "Function code", "field_type": "FIXED", "field_length": 3},
        "25": {"field_name": "Message reason code", "field_type": "FIXED", "field_length": 4},
        "26": {"field_name": "Card acceptor business code", "field_type": "FIXED", "field_length": 4,
               "field_python_type": "int"},
        "30": {"field_name": "Amounts, original", "field_type": "FIXED", "field_length": 24},
        "31": {"field_name": "Acquirer reference data", "field_type": "LLVAR", "field_length": 23},
        "32": {"field_name": "Acquiring institution ID code", "field_type": "LLVAR", "field_length": 0},
        "33": {"field_name": "Forwarding institution ID code", "field_type": "LLVAR", "field_length": 0},
        "37": {"field_name": "Retrieval reference number", "field_type": "FIXED", "field_length": 12},
        "38": {"field_name": "Approval code", "field_type": "FIXED", "field_length": 6},
        "40": {"field_name": "Service code", "field_type": "FIXED", "field_length": 3},
        "41": {"field_name": "Card acceptor terminal ID", "field_type": "FIXED", "field_length": 8},
        "42": {"field_name": "Card acceptor Id", "field_type": "FIXED", "field_length": 15},
        "43": {"field_name": "Card acceptor name/location", "field_type": "LLVAR", "field_length": 0,
               "field_processor": "DE43"},
        "48": {"field_name": "Additional data", "field_type": "LLLVAR", "field_length": 0, "field_processor": "PDS"},
        "49": {"field_name": "Currency code, Transaction", "field_type": "FIXED", "field_length": 3},
        "50": {"field_name": "Currency code, Reconciliation", "field_type": "FIXED", "field_length": 3},
        "51": {"field_name": "Currency code, Cardholder billing", "field_type": "FIXED", "field_length": 3},
        "54": {"field_name": "Amounts, additional", "field_type": "LLLVAR", "field_length": 0},
        "55": {"field_name": "ICC system related data", "field_type": "LLLVAR", "field_length": 255,
               "field_processor": "ICC"},
        "62": {"field_name": "Additional data 2", "field_type": "LLLVAR", "field_length": 0, "field_processor": "PDS"},
        "63": {"field_name": "Transaction lifecycle Id", "field_type": "LLLVAR", "field_length": 16},
        "71": {"field_name": "Message number", "field_type": "FIXED", "field_length": 8, "field_python_type": "int"},
        "72": {"field_name": "Data record", "field_type": "LLLVAR", "field_length": 0},
        "73": {"field_name": "Date, Action", "field_type": "FIXED", "field_length": 6},
        "93": {"field_name": "Transaction destination institution ID", "field_type": "LLVAR", "field_length": 0},
        "94": {"field_name": "Transaction originator institution ID", "field_type": "LLVAR", "field_length": 0},
        "95": {"field_name": "Card issuer reference data", "field_type": "LLVAR", "field_length": 10},
        "100": {"field_name": "Receiving institution ID", "field_type": "LLVAR", "field_length": 11},
        "111": {"field_name": "Amount, currency conversion assignment", "field_type": "LLLVAR", "field_length": 0},
        "123": {"field_name": "Additional data 3", "field_type": "LLLVAR", "field_length": 0, "field_processor": "PDS"},
        "124": {"field_name": "Additional data 4", "field_type": "LLLVAR", "field_length": 0, "field_processor": "PDS"},
        "125": {"field_name": "Additional data 5", "field_type": "LLLVAR", "field_length": 0, "field_processor": "PDS"},
        "127": {"field_name": "Network data", "field_type": "LLLVAR", "field_length": 0}},

    "output_data_elements": [
        "MTI", "DE2", "DE3", "DE4", "DE12", "DE14", "DE22", "DE23", "DE24", "DE25", "DE26",
        "DE30", "DE31", "DE33", "DE37", "DE38", "DE40", "DE41", "DE42", "DE48", "DE49",
        "DE50", "DE63", "DE71", "DE73", "DE93", "DE94", "DE95", "DE100", "PDS0023",
        "PDS0052", "PDS0122", "PDS0148", "PDS0158", "PDS0165", "DE43_NAME", "DE43_SUBURB",
        "DE43_POSTCODE", "ICC_DATA"
    ]
}



import csv
import sys


def dicts_to_csv(data_list, field_list, output_file):
    """
    Writes dict data to CSV file

    :param data_list: list of dictionaries that contain the data to be loaded
    :param field_list: list of fields in the dictionary to be loaded
    :param output_filename: filename for output CSV file
    :return: None
    """
    try:
        instance_type = unicode
        file_mode = "wb"
    except NameError:
        instance_type = str
        file_mode = "w"

    filtered_data_list = filter_data_list(data_list, field_list)

    # with open(output_filename, file_mode) as output_file:
    writer = csv.DictWriter(
        output_file,
        fieldnames=field_list,
        extrasaction="ignore",
        lineterminator="\n")

    # python 2.6 does not support writeheader() so skip
    if sys.version_info[0] == 2 and sys.version_info[1] == 6:
        pass
    else:
        writer.writeheader()

    for item in filtered_data_list:
        if file_mode == "w":
            row = dict(
                (k, v.decode('latin1') if not isinstance(v, instance_type) else v)
                for k, v in item.items()
            )
        else:
            row = dict(
                (k, v.encode('utf-8') if isinstance(v, instance_type) else v)
                for k, v in item.items()
            )
        writer.writerow(row)

    # LOGGER.info("%s records written", len(data_list))


def filter_data_list(data_list, field_list):
    """
    Takes list of dictionaries and returns new list filtered to only
    return the field keys provided in field_list and the values
    decoded to unicode.

    :param data_list: the list of dictionaries
    :param field_list: the list of string keys to return
    :return: filtered data list
    """
    return [filter_dictionary(item, field_list) for item in data_list]


def filter_dictionary(dictionary, field_list):
    """
    Takes dictionary and list of elements and returns dictionary with just
    elements specified. Also decodes the items to unicode

    :param dictionary: the dictionary to filter
    :param field_list: list containing keys to keep
    :return: dictionary with just keys from list. All values decoded
    """
    return {item: dictionary[item] for item in dictionary if item in field_list}



"""
MasterCard file formats
-----------------------

VBS file format
^^^^^^^^^^^^^^^

This format is a basic variable record format.

There are no carriage returns or line feeds in the file.
A file consists of records. Each record is prefixed with a 4 byte binary length.

Say you had a file with the following 2 records:

.. code-block:: none

    "This is first record 1234567"  <- length 28
    "This is second record AAAABBBBB123"  <- length 34

Add 4 byte binary length to the start of each record. (x'1C' = 28, x'22' = 34)
with the file finishing with a zero length record length

.. code-block:: hexdump

    00000000: 00 00 00 1C 54 68 69 73  20 69 73 20 66 69 72 73  ....This is firs
    00000010: 74 20 72 65 63 6F 72 64  20 31 32 33 34 35 36 37  t record 1234567
    00000020: 00 00 00 22 54 68 69 73  20 69 73 20 73 65 63 6F  ..."This is seco
    00000030: 6E 64 20 72 65 63 6F 72  64 20 41 41 41 41 42 42  nd record AAAABB
    00000040: 42 42 42 31 32 33 00 00  00 00                    BBB123....

1014 blocked file format
^^^^^^^^^^^^^^^^^^^^^^^^
This is the same as VBS format with 1014 blocking applied.

The VBS data is blocked into lengths of 1012, and an additional
2 x'40' characters are appended at each block.

Finally, the total file length is made a multiple of 1014 with the final
incomplete record being filled with the x'40' character

Taking the above VBS example

.. code-block:: hexdump

    00000000: 00 00 00 1C 54 68 69 73  20 69 73 20 66 69 72 73  ....This is firs
    00000010: 74 20 72 65 63 6F 72 64  20 31 32 33 34 35 36 37  t record 1234567
    00000020: 00 00 00 22 54 68 69 73  20 69 73 20 73 65 63 6F  ..."This is seco
    00000030: 6E 64 20 72 65 63 6F 72  64 20 41 41 41 41 42 42  nd record AAAABB
    00000040: 42 42 42 31 32 33 00 00  00 00                    BBB123....

Block to 1014 by adding 2 * x'40' characters every 1012 characters in the data.
Finally fill with x'40' characters to next 1014 increment.
In this case, there is only one increment

.. code-block:: hexdump

    00000000: 00 00 00 1C 54 68 69 73  20 69 73 20 66 69 72 73  ....This is firs
    00000010: 74 20 72 65 63 6F 72 64  20 31 32 33 34 35 36 37  t record 1234567
    00000020: 00 00 00 22 54 68 69 73  20 69 73 20 73 65 63 6F  ..."This is seco
    00000030: 6E 64 20 72 65 63 6F 72  64 20 41 41 41 41 42 42  nd record AAAABB
    00000040: 42 42 42 31 32 33 00 00  00 00 40 40 40 40 40 40  BBB123....@@@@@@
    00000050: 40 40 40 40 40 40 40 40  40 40 40 40 40 40 40 40  @@@@@@@@@@@@@@@@
    000003E0: 40 40 40 40 40 40 40 40  40 40 40 40 40 40 40 40  @@@@@@@@@@@@@@@@
    000003F0: 40 40 40 40 40 40                                 @@@@@@

"""

import logging
import struct

from cardutil import iso8583
from cardutil.config import config
from cardutil.outputter import dicts_to_csv

LOGGER = logging.getLogger(__name__)


def change_encoding(in_file, out_file, in_encoding='cp500', out_encoding='latin-1'):
    """
    Change encoding of IPM file from one encoding scheme to another

    :param in_file: input IPM file object
    :param out_file: output IPM file object
    :param in_encoding: input file encoding string
    :param out_encoding: output file encoding string
    :return: None
    """
    reader = IpmReader(in_file, encoding=in_encoding, blocked=True)
    writer = IpmWriter(out_file, encoding=out_encoding, blocked=True)

    for record in reader:
        writer.write(record)
    writer.close()  # finalises the file by adding zero length record


def change_param_encoding(in_file, out_file, in_encoding='cp500', out_encoding='latin-1'):
    """
    Change encoding of parameter file from one encoding format to another.

    :param in_file: input parameter file object
    :param out_file: output parameter file object
    :param in_encoding: input file encoding string
    :param out_encoding: output file encoding string
    :return: None
    """
    vbs_reader = VbsReader(in_file, blocked=True)

    in_records = (record.decode(in_encoding) for record in vbs_reader)
    out_records = (record.encode(out_encoding) for record in in_records)

    vbs_writer = VbsWriter(out_file, blocked=True)
    for record in out_records:
        vbs_writer.write(record)
    vbs_writer.close()


def output_csv(in_file, out_file, in_encoding='latin-1'):
    """
    Extract IPM file to csv format

    :param in_file: input IPM file object
    :param out_file: output csv file object
    :param in_encoding: input file encoding string
    :return: None
    """
    dicts_to_csv(IpmReader(Unblock1014(in_file), encoding=in_encoding), config['output_data_elements'], out_file)


class Block1014(object):
    """
    1014 Blocker for file objects.
    Wrap around a file object. Return 1014 blocked data
    """
    PAD_CHAR = b'\x40'

    def __init__(self, file_obj):
        self.file_obj = file_obj
        self.remaining_chars = 1012

    def __getattr__(self, name):
        """
        Attribute proxy for wrapped file object
        """
        try:
            return self.__dict__[name]
        except KeyError:
            if hasattr(self.file_obj, name):
                return getattr(self.file_obj, name)
        return None

    def write(self, bytes_to_write):
        """
        Write requested bytes to the output file object.
        """
        # not enough bytes to complete a block, just write and subtract from remaining
        LOGGER.debug(f'bytes_to_write={bytes_to_write}')
        if len(bytes_to_write) < self.remaining_chars:
            LOGGER.debug(f'len->{len(bytes_to_write)}<{self.remaining_chars}')
            LOGGER.debug(f'write {bytes_to_write}')
            self.file_obj.write(bytes_to_write)
            self.remaining_chars -= len(bytes_to_write)
            return

        # complete the first record
        LOGGER.debug(f'write first: {bytes_to_write[:self.remaining_chars]}')
        self.file_obj.write(bytes_to_write[:self.remaining_chars])
        self.file_obj.write(self.PAD_CHAR * 2)
        bytes_to_write = bytes_to_write[self.remaining_chars:]

        # now write complete blocks
        while len(bytes_to_write) > 1012:
            LOGGER.debug(f'write while: {bytes_to_write[:1012]}')
            self.file_obj.write(bytes_to_write[:1012])
            self.file_obj.write(self.PAD_CHAR * 2)
            bytes_to_write = bytes_to_write[1012:]

        # write whatever is left
        LOGGER.debug(f'write last: {bytes_to_write}')
        self.file_obj.write(bytes_to_write)
        self.remaining_chars = 1014-len(bytes_to_write)
        LOGGER.debug(f'remaining_chars={self.remaining_chars}')

    def seek(self, pos):
        """
        Finalise then seek file object to requested position

        .. note:: Method only partially implemented. Only use to seek start of file (zero)
        """
        self.finalise()
        self.file_obj.seek(pos)

    def close(self):
        """
        Finalise then close the file object
        """
        self.finalise()
        self.file_obj.close()

    def finalise(self):
        """
        Complete the blocking operation by creating final 1014 block.
        Called by ``close`` and ``seek`` methods to ensure completion.
        """
        self.file_obj.write(self.PAD_CHAR * (self.remaining_chars + 2))
        self.remaining_chars = 1012


class Unblock1014(object):
    """
    Unblocks 1014 blocked file objects.
    Wrap around a 1014 blocked file object. Return file like object providing only unblocked data
    """
    def __init__(self, file_obj):
        self.file_obj = file_obj
        self.buffer = b''

    def __getattr__(self, name):
        """
        Attribute proxy for wrapped file object
        """
        try:
            return self.__dict__[name]
        except KeyError:
            if hasattr(self.file_obj, name):
                return getattr(self.file_obj, name)
        return None

    def read(self, bytes_to_read=0):
        """
        Read requested bytes from the file object. Returned data will be unblocked
        """
        read_all = True if not bytes_to_read else False
        while read_all or len(self.buffer) <= bytes_to_read:
            block = self.file_obj.read(1014)
            if not block:  # eof
                break
            self.buffer += block[:1012]
        output = self.buffer[:bytes_to_read]
        self.buffer = self.buffer[bytes_to_read:]
        return output


class VbsReader(object):
    """
    The VbsReader class can be used to iterate through a VBS formatted file
    object record by record.

    ::

        from cardutil.mciipm import VbsReader
        with open('vbs_file.bin', 'rb') as vbs_file:
            vbs_reader = VbsReader(vbs_file)
            for vbs_record in vbs_reader:
                print(vbs_record)

    """
    def __init__(self, vbs_file, blocked=False):
        """
        Create a VbsReader object

        :param vbs_file: File object with VBS formatted data
        """
        self.vbs_data = vbs_file
        if blocked:
            self.vbs_data = Unblock1014(vbs_file)

    def __getattr__(self, name):
        """
        Attribute proxy for wrapped file object
        """
        try:
            return self.__dict__[name]
        except KeyError:
            if hasattr(self.vbs_data, name):
                return getattr(self.vbs_data, name)
        return None

    def __iter__(self):
        return self

    def __next__(self):
        """
        Unpacks a variable blocked object into records
        """
        record_length_raw = self.vbs_data.read(4)
        record_length = struct.unpack(">i", record_length_raw)[0]
        LOGGER.debug("record_length=%s", record_length)

        # exit if last record (length=0)
        if record_length == 0:
            raise StopIteration

        record = self.vbs_data.read(record_length)
        return record


class IpmReader(VbsReader):
    """
    IPM reader can be used to iterate through an IPM file

    The file object must be in VBS format.

    ::

        from cardutil.mciipm import IpmReader
        with open('vbs_in.bin', 'rb') as vbs_in:
            reader = IpmReader(vbs_in)
            for record in reader:
                print(record)

    If the file required 1014 block format, then set the ``blocked`` parameter to True.

    ::

        from cardutil.mciipm import IpmReader
        with open('blocked_in.bin', 'rb') as blocked_in:
            reader = IpmReader(blocked_in, blocked=True)
            for record in reader:
                print(record)

    """
    def __init__(self, ipm_file, encoding='ascii', **kwargs):
        """
        Create a new IpmReader

        :param ipm_file: the file object to read
        :param encoding: the file encoding
        """
        self.encoding = encoding
        super(IpmReader, self).__init__(ipm_file, **kwargs)

    def __next__(self):
        vbs_record = super(IpmReader, self).__next__()
        LOGGER.debug(f'{len(vbs_record)}: {vbs_record}')
        return iso8583.loads(vbs_record, encoding=self.encoding)


class VbsWriter(object):
    """
    Writes VBS formatted files.

    The writer can be used as follows::

        with open('vbs_out.bin', 'wb') as vbs_out:
            writer = VbsWriter(vbs_out)
            writer.write(b'This is the record')
            writer.close()

    The message is a byte string containing the data.

    The `close` method must be issued to finalise the file by adding the zero length record which
    indicated the end of the file.
    """
    def __init__(self, out_file, blocked=False):
        self.out_file = out_file
        if blocked:
            self.out_file = Block1014(out_file)

    def __getattr__(self, name):
        """
        Attribute proxy for wrapped file object
        """
        try:
            return self.__dict__[name]
        except KeyError:
            if hasattr(self.out_file, name):
                return getattr(self.out_file, name)
        return None

    def write(self, record):
        """
        Add a new record to the VBS output file

        :param record: byte string containing data
        :return: None
        """
        # get the length of the record
        record_length = len(record)
        # convert length to binary
        record_length_raw = struct.pack(">i", record_length)
        # add length to output data
        self.out_file.write(record_length_raw)
        # add data to output
        self.out_file.write(record)

    def close(self):
        """
        Finalise the VBS file output by adding the zero length file record.

        :return: None
        """
        # add zero length to end of record
        self.out_file.write(struct.pack(">i", 0))
        self.out_file.seek(0)


class IpmWriter(VbsWriter):
    """
    IPM writer can be used to write records to a Mastercard IPM file

    ::

        from cardutil.mciipm import IpmWriter
        with open('ipm_out.bin', 'wb') as ipm_out:
            writer = IpmWriter(ipm_out)
            writer.write({'MTI': '1111', 'DE2': '9999111122221111'})
            writer.close()

    If the required file is 1014 block format, then set the ``blocked`` parameter to True.

    ::

        from cardutil.mciipm import IpmWriter
        with open('ipm_out.bin', 'wb') as ipm_out:
            writer = IpmWriter(ipm_out, blocked=True)
            writer.write({'MTI': '1111', 'DE2': '9999111122221111'})
            writer.close()

    By default, the file output is ``ascii`` encoded, but you can provide an alternative
    encoding if required. All standard python encoding schemes are supported.
    Mainframe systems likely use ``cp500``

    ::

       writer = IpmWriter(vbs_out, encoding='cp500')
       writer.write({'MTI': '1111', 'DE2': '9999111122221111'})
       writer.close()

    """
    def __init__(self, file_obj, encoding='ascii', **kwargs):
        """
        Create a new IpmWriter

        :param file_obj: the file object to write to
        :param encoding: the file encoding
        """
        self.encoding = encoding
        super(IpmWriter, self).__init__(file_obj, **kwargs)

    def write(self, obj):
        """
        Writes new record to IPM file

        :param obj: dictionary object containing ISO8583 elements.

        See :py:mod:`cardutil.iso8583` for expected dict object keys.

        :return: None
        """
        record = iso8583.dumps(obj, encoding=self.encoding)
        super(IpmWriter, self).write(record)


def unblock_1014(input_data, output_data):
    """
    Unblocks a 1014 byte blocked file object

    :param input_data: 1014 blocked IPM file object
    :param output_data: unblocked file object
    """
    pad_char = b'\x40'
    while True:
        record = input_data.read(1014)
        if not record:
            break
        if len(record) != 1014:
            raise ValueError('Invalid file size')
        if record[-2:] != pad_char * 2:
            raise ValueError('Invalid 1014 block line ending')
        output_data.write(record[0:1012])
    output_data.seek(0)
    input_data.seek(0)


def block_1014(input_data, output_data):
    """
    Creates a 1014 byte blocked file object

    :param input_data: file object to be 1014 byte blocked
    :param output_data: 1014 byte blocked file object
    """
    pad_char = b'\x40'

    while True:
        record = input_data.read(1012)
        # end of data
        if not record:
            break
        # incomplete 1012 block
        if len(record) != 1012:
            record += (1012 - len(record)) * pad_char
        output_data.write(record + (pad_char * 2))
    output_data.seek(0)
    input_data.seek(0)



from setuptools import setup, find_packages

readme = open('README.rst', 'r').read()

setup(name='cardutil',
      packages=find_packages(),
      version='0.1.4',
      license='MIT',
      description='Python package for working with payment card systems',
      long_description=readme,
      long_description_content_type='text/x-rst',
      author='Anthony Delosa',
      author_email='adelosa@gmail.com',
      url='https://bitbucket.org/hoganman/cardutil',
      install_requires=[],
      extras_require={
        'docs': ['sphinx', 'sphinx_rtd_theme'],
        'test': ['flake8', 'pytest', 'bumpversion']
      },
      classifiers=[
            'Development Status :: 5 - Production/Stable',
            # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
            'License :: OSI Approved :: MIT License',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
      ])
